import logging
import random

import requests
from requests import RequestException

from bubuku.zookeeper import AddressListProvider

_LOG = logging.getLogger('bubuku.zookeeper.exhibitor')


class ExhibitorAddressProvider(AddressListProvider):
    def __init__(self, initial_list_provider):
        self.initial_list_provider = initial_list_provider
        self.exhibitors = []

    def get_latest_address(self) -> (list, int):
        json_ = self._query_exhibitors(self.exhibitors)
        if not json_:
            self.exhibitors = self.initial_list_provider()
            json_ = self._query_exhibitors(self.exhibitors)
        if isinstance(json_, dict) and 'servers' in json_ and 'port' in json_:
            self.exhibitors = json_['servers']
            return sorted(json_['servers']), int(json_['port'])
        return None

    def _query_exhibitors(self, exhibitors):
        if not exhibitors:
            return None
        random.shuffle(exhibitors)
        for host in exhibitors:
            url = 'http://{}:{}{}'.format(host, 8181, '/exhibitor/v1/cluster/list')
            try:
                response = requests.get(url, timeout=3.1)
                return response.json()
            except RequestException as e:
                _LOG.warn('Failed to query zookeeper list information from {}'.format(url), exc_info=e)
        return None



import logging

from bubuku.broker import BrokerManager
from bubuku.controller import Change, Check
from bubuku.zookeeper import BukuExhibitor

_LOG = logging.getLogger('bubuku.features.restart_on_zk')

_STAGE_STOP = 'stop'
_STAGE_START = 'start'
_STAGE_WAIT_FOR_ISR = 'wait_for_isr'


class RestartBrokerChange(Change):
    def __init__(self, zk: BukuExhibitor, broker: BrokerManager, break_condition, processed_callback=None):
        self.zk = zk
        self.broker = broker
        self.break_condition = break_condition
        self.stage = _STAGE_STOP
        self.processed_callback = processed_callback

    def get_name(self):
        return 'restart'

    def can_run(self, current_actions):
        return all([a not in current_actions for a in ['start', 'restart', 'stop']])

    def run(self, current_actions):
        if self.stage == _STAGE_STOP:
            if self.break_condition and self.break_condition():
                return False
            self.broker.stop_kafka_process()
            self.stage = _STAGE_START
            return True
        elif self.stage == _STAGE_START:
            # Yep, use latest data
            zk_conn_str = self.zk.get_conn_str()
            try:
                self.broker.start_kafka_process(zk_conn_str)
            except Exception as e:
                _LOG.error('Failed to start kafka process against {}'.format(zk_conn_str), exc_info=e)
                return True
            self.stage = _STAGE_WAIT_FOR_ISR
            return True
        elif self.stage == _STAGE_WAIT_FOR_ISR:
            if not self.broker.is_running_and_registered():
                _LOG.warning("Broker is considered to be dead right after restart, won't wait for isr")
                return False
            unjoined = self.broker.get_disjoined_isr_topic_partitions()
            if unjoined:
                _LOG.info("Still waiting to join to isr list for topic partitions: {}".format(unjoined))
                return True
            return False
        else:
            _LOG.error('Stage {} is not supported'.format(self.stage))
        return False

    def on_remove(self):
        if self.processed_callback:
            self.processed_callback()

    def __str__(self):
        return 'RestartOnZkChange ({}), stage={}'.format(self.get_name(), self.stage)


class CheckExhibitorAddressChanged(Check):
    def __init__(self, zk: BukuExhibitor, broker: BrokerManager):
        super().__init__()
        self.zk = zk
        self.broker = broker
        self.conn_str = None

    def check(self) -> Change:
        new_conn_str = self.zk.get_conn_str()
        if new_conn_str != self.conn_str:
            def _cancel_if():
                current_conn_str = self.zk.get_conn_str()
                if current_conn_str != new_conn_str:
                    _LOG.warning('ZK address changed again, from {} to {}'.format(new_conn_str, current_conn_str))
                    return True
                if current_conn_str == self.broker.get_zk_connect_string():
                    _LOG.warning('Broker already have latest version of zk address: '.format(current_conn_str))
                    return True
                return False

            _LOG.info('ZK addresses changed from {} to {}, triggering restart'.format(self.conn_str, new_conn_str))
            self.conn_str = new_conn_str
            return RestartBrokerChange(self.zk, self.broker, _cancel_if)

    def __str__(self):
        return 'CheckExhibitorAddressChanged, current={}'.format(self.conn_str)



#!/usr/bin/env python3
import functools
import logging
import os
import re
from time import sleep, time

from bubuku.config import KafkaProperties
from bubuku.zookeeper import BukuExhibitor

_LOG = logging.getLogger('bubuku.id_generator')


class BrokerIdGenerator(object):
    def get_broker_id(self) -> str:
        raise NotImplementedError('Not implemented')

    def detect_broker_id(self):
        raise NotImplementedError('Not implemented')

    def wait_for_broker_id_absence(self):
        while self.is_registered():
            sleep(1)

    def wait_for_broker_id_presence(self, timeout) -> bool:
        start = time()
        while not self.is_registered():
            if (time() - start) > timeout:
                return False
            sleep(1)
        return True

    def is_registered(self):
        raise NotImplementedError('Not implemented')


def _create_rfc1918_address_hash(ip: str) -> (str, str):
    address = [int(v) for v in ip.split('.')]
    # the goal of this hashing is to get positive 4-bytes int which can not be changed during restarts
    if address[0] == 10:
        address[0] = 1
    elif address[0] == 192 and address[1] == 168:
        address[0] = 2
    elif address[0] == 172 and (address[1] & 0xF0) == 16:
        address[0] = 3
    else:
        return None
    return str(functools.reduce(lambda o, v: o * 256 + v, address, 0)), str(256 * 256 * 256 * 4 + 1)


class BrokerIDByIp(BrokerIdGenerator):
    def __init__(self, zk: BukuExhibitor, ip: str, kafka_props: KafkaProperties):
        self.zk = zk
        self.broker_id, max_id = _create_rfc1918_address_hash(ip)
        kafka_props.set_property('reserved.broker.max.id', max_id)
        _LOG.info('Built broker id {} from ip: {}'.format(self.broker_id, ip))
        if self.broker_id is None:
            raise NotImplementedError('Broker id from ip address supported only for rfc1918 private addresses')

    def get_broker_id(self):
        return self.broker_id

    def detect_broker_id(self):
        return self.broker_id

    def is_registered(self):
        return self.zk.is_broker_registered(self.broker_id)


class BrokerIdAutoAssign(BrokerIdGenerator):
    def __init__(self, zk: BukuExhibitor, kafka_properties: KafkaProperties):
        super().__init__()
        self.zk = zk
        self.kafka_properties = kafka_properties
        self.broker_id = None

    def get_broker_id(self):
        return None

    def detect_broker_id(self):
        meta_path = '{}/meta.properties'.format(self.kafka_properties.get_property('log.dirs'))
        while not os.path.isfile(meta_path):
            return None
        with open(meta_path) as f:
            lines = f.readlines()
            for line in lines:
                match = re.search('broker\.id=(\d+)', line)
                if match:
                    return match.group(1)
        return None

    def is_registered(self):
        broker_id = self.detect_broker_id()
        if broker_id:
            return self.zk.is_broker_registered(broker_id)
        return False
